package ru.dbotthepony.mc.otm.registry

import net.minecraft.core.BlockPos
import net.minecraft.resources.ResourceLocation
import net.minecraft.world.damagesource.DamageSource
import net.minecraft.world.entity.EntityType
import net.minecraft.world.item.BlockItem
import net.minecraft.world.item.DyeColor
import net.minecraft.world.item.Item
import net.minecraft.world.level.BlockGetter
import net.minecraft.world.level.block.*
import net.minecraft.world.level.block.state.BlockBehaviour
import net.minecraft.world.level.block.state.BlockState
import net.minecraft.world.level.material.Material
import net.minecraft.world.level.material.MaterialColor
import net.minecraftforge.fml.javafmlmod.FMLJavaModLoadingContext
import net.minecraftforge.registries.DeferredRegister
import net.minecraftforge.registries.ForgeRegistry
import net.minecraftforge.registries.IForgeRegistry
import net.minecraftforge.registries.NewRegistryEvent
import net.minecraftforge.registries.RegistryBuilder
import net.minecraftforge.registries.RegistryObject
import ru.dbotthepony.mc.otm.OverdriveThatMatters
import ru.dbotthepony.mc.otm.android.AndroidFeatureType
import ru.dbotthepony.mc.otm.android.AndroidResearchType
import java.util.function.Supplier
import kotlin.properties.Delegates
import kotlin.properties.ReadOnlyProperty
import kotlin.properties.ReadWriteProperty
import kotlin.reflect.KProperty

class CrateProperties(val color: MaterialColor, val name: String) {
	fun makeBlock(): Block {
		return Block(BlockBehaviour.Properties.of(Material.METAL, color)
			.sound(SoundType.METAL)
			.requiresCorrectToolForDrops()
			.strength(5.0F, 6.0F))
	}
}

class IndustrialGlassProperties(val color: DyeColor?, val name: String, val namePane: String) {
	constructor(color: DyeColor?) : this(color, "industrial_glass${color?.let { "_$it" } ?: ""}", "industrial_glass_pane${color?.let { "_$it" } ?: ""}")

	fun makeBlock(): Block {
		val properties = BlockBehaviour.Properties.of(Material.GLASS, if (color != null) color.materialColor else MaterialColor.NONE)
				.strength(1.5f, 35.0f)
				.requiresCorrectToolForDrops()
				.sound(SoundType.GLASS)
				.noOcclusion()
				.isValidSpawn { _: BlockState, _: BlockGetter, _: BlockPos, _: EntityType<*>? -> false }
				.isRedstoneConductor { _: BlockState, _: BlockGetter, _: BlockPos -> false }
				.isSuffocating { _: BlockState, _: BlockGetter, _: BlockPos -> false }
				.isViewBlocking { _: BlockState, _: BlockGetter, _: BlockPos -> false }

		if (color != null) {
			return StainedGlassBlock(color, properties)
		}

		return GlassBlock(properties)
	}

	fun makePaneBlock(): Block {
		val properties = BlockBehaviour.Properties.of(Material.GLASS, if (color != null) color.materialColor else MaterialColor.NONE)
			.strength(1.25f, 5.0f)
			.requiresCorrectToolForDrops()
			.sound(SoundType.GLASS)
			.noOcclusion()

		if (color != null) {
			return StainedGlassPaneBlock(color, properties)
		}

		return IronBarsBlock(properties)
	}
}

private class RegistryDelegate<V>(private val name: String) {
	private var value: Supplier<V>? = null

	fun get(): V {
		val supp = value ?: throw IllegalStateException("Tried to access uninitialized registry $name")
		return supp.get() ?: throw IllegalStateException("Accessing registry $name too early")
	}

	fun write(value: Supplier<V>) {
		if (this.value != null) {
			throw IllegalStateException("Already have registry written")
		}

		this.value = value
	}
}

private class WriteOnce<V> : ReadWriteProperty<Any, V> {
	private var value: V? = null

	override fun getValue(thisRef: Any, property: KProperty<*>): V {
		return checkNotNull(value) { "Property wasn't initialized" }
	}

	override fun setValue(thisRef: Any, property: KProperty<*>, value: V) {
		if (this.value != null) {
			throw IllegalStateException("Property already initialized")
		}

		this.value = value
	}
}

class ColoredDecorativeBlock(
	val baseName: String,
	private val propertySupplier: (DyeColor) -> BlockBehaviour.Properties
) {
	private var _whiteBlock: RegistryObject<Block> by WriteOnce()
	private var _orangeBlock: RegistryObject<Block> by WriteOnce()
	private var _magentaBlock: RegistryObject<Block> by WriteOnce()
	private var _lightBlueBlock: RegistryObject<Block> by WriteOnce()
	private var _yellowBlock: RegistryObject<Block> by WriteOnce()
	private var _limeBlock: RegistryObject<Block> by WriteOnce()
	private var _pinkBlock: RegistryObject<Block> by WriteOnce()
	private var _grayBlock: RegistryObject<Block> by WriteOnce()
	private var _lightGrayBlock: RegistryObject<Block> by WriteOnce()
	private var _cyanBlock: RegistryObject<Block> by WriteOnce()
	private var _purpleBlock: RegistryObject<Block> by WriteOnce()
	private var _blueBlock: RegistryObject<Block> by WriteOnce()
	private var _brownBlock: RegistryObject<Block> by WriteOnce()
	private var _greenBlock: RegistryObject<Block> by WriteOnce()
	private var _redBlock: RegistryObject<Block> by WriteOnce()
	private var _blackBlock: RegistryObject<Block> by WriteOnce()

	val whiteBlock: Block get() = _whiteBlock.get()
	val orangeBlock: Block get() = _orangeBlock.get()
	val magentaBlock: Block get() = _magentaBlock.get()
	val lightBlueBlock: Block get() = _lightBlueBlock.get()
	val yellowBlock: Block get() = _yellowBlock.get()
	val limeBlock: Block get() = _limeBlock.get()
	val pinkBlock: Block get() = _pinkBlock.get()
	val grayBlock: Block get() = _grayBlock.get()
	val lightGrayBlock: Block get() = _lightGrayBlock.get()
	val cyanBlock: Block get() = _cyanBlock.get()
	val purpleBlock: Block get() = _purpleBlock.get()
	val blueBlock: Block get() = _blueBlock.get()
	val brownBlock: Block get() = _brownBlock.get()
	val greenBlock: Block get() = _greenBlock.get()
	val redBlock: Block get() = _redBlock.get()
	val blackBlock: Block get() = _blackBlock.get()

	private var _whiteItem: RegistryObject<Item> by WriteOnce()
	private var _orangeItem: RegistryObject<Item> by WriteOnce()
	private var _magentaItem: RegistryObject<Item> by WriteOnce()
	private var _lightBlueItem: RegistryObject<Item> by WriteOnce()
	private var _yellowItem: RegistryObject<Item> by WriteOnce()
	private var _limeItem: RegistryObject<Item> by WriteOnce()
	private var _pinkItem: RegistryObject<Item> by WriteOnce()
	private var _grayItem: RegistryObject<Item> by WriteOnce()
	private var _lightGrayItem: RegistryObject<Item> by WriteOnce()
	private var _cyanItem: RegistryObject<Item> by WriteOnce()
	private var _purpleItem: RegistryObject<Item> by WriteOnce()
	private var _blueItem: RegistryObject<Item> by WriteOnce()
	private var _brownItem: RegistryObject<Item> by WriteOnce()
	private var _greenItem: RegistryObject<Item> by WriteOnce()
	private var _redItem: RegistryObject<Item> by WriteOnce()
	private var _blackItem: RegistryObject<Item> by WriteOnce()

	val whiteItem: Item get() = _whiteItem.get()
	val orangeItem: Item get() = _orangeItem.get()
	val magentaItem: Item get() = _magentaItem.get()
	val lightBlueItem: Item get() = _lightBlueItem.get()
	val yellowItem: Item get() = _yellowItem.get()
	val limeItem: Item get() = _limeItem.get()
	val pinkItem: Item get() = _pinkItem.get()
	val grayItem: Item get() = _grayItem.get()
	val lightGrayItem: Item get() = _lightGrayItem.get()
	val cyanItem: Item get() = _cyanItem.get()
	val purpleItem: Item get() = _purpleItem.get()
	val blueItem: Item get() = _blueItem.get()
	val brownItem: Item get() = _brownItem.get()
	val greenItem: Item get() = _greenItem.get()
	val redItem: Item get() = _redItem.get()
	val blackItem: Item get() = _blackItem.get()

	val blocks = LazyList(
		{ _whiteBlock.get() },
		{ _orangeBlock.get() },
		{ _magentaBlock.get() },
		{ _lightBlueBlock.get() },
		{ _yellowBlock.get() },
		{ _limeBlock.get() },
		{ _pinkBlock.get() },
		{ _grayBlock.get() },
		{ _lightGrayBlock.get() },
		{ _cyanBlock.get() },
		{ _purpleBlock.get() },
		{ _blueBlock.get() },
		{ _brownBlock.get() },
		{ _greenBlock.get() },
		{ _redBlock.get() },
		{ _blackBlock.get() },
	)

	val items = LazyList(
		{ _whiteItem.get() },
		{ _orangeItem.get() },
		{ _magentaItem.get() },
		{ _lightBlueItem.get() },
		{ _yellowItem.get() },
		{ _limeItem.get() },
		{ _pinkItem.get() },
		{ _grayItem.get() },
		{ _lightGrayItem.get() },
		{ _cyanItem.get() },
		{ _purpleItem.get() },
		{ _blueItem.get() },
		{ _brownItem.get() },
		{ _greenItem.get() },
		{ _redItem.get() },
		{ _blackItem.get() },
	)

	fun registerBlocks(registry: DeferredRegister<Block>) {
		_whiteBlock = registry.register("${baseName}_white") { Block(propertySupplier.invoke(DyeColor.WHITE)) }
		_orangeBlock = registry.register("${baseName}_orange") { Block(propertySupplier.invoke(DyeColor.ORANGE)) }
		_magentaBlock = registry.register("${baseName}_magenta") { Block(propertySupplier.invoke(DyeColor.MAGENTA)) }
		_lightBlueBlock = registry.register("${baseName}_light_blue") { Block(propertySupplier.invoke(DyeColor.LIGHT_BLUE)) }
		_yellowBlock = registry.register("${baseName}_yellow") { Block(propertySupplier.invoke(DyeColor.YELLOW)) }
		_limeBlock = registry.register("${baseName}_lime") { Block(propertySupplier.invoke(DyeColor.LIME)) }
		_pinkBlock = registry.register("${baseName}_pink") { Block(propertySupplier.invoke(DyeColor.PINK)) }
		_grayBlock = registry.register("${baseName}_gray") { Block(propertySupplier.invoke(DyeColor.GRAY)) }
		_lightGrayBlock = registry.register("${baseName}_light_gray") { Block(propertySupplier.invoke(DyeColor.LIGHT_GRAY)) }
		_cyanBlock = registry.register("${baseName}_cyan") { Block(propertySupplier.invoke(DyeColor.CYAN)) }
		_purpleBlock = registry.register("${baseName}_purple") { Block(propertySupplier.invoke(DyeColor.PURPLE)) }
		_blueBlock = registry.register("${baseName}_blue") { Block(propertySupplier.invoke(DyeColor.BLUE)) }
		_brownBlock = registry.register("${baseName}_brown") { Block(propertySupplier.invoke(DyeColor.BROWN)) }
		_greenBlock = registry.register("${baseName}_green") { Block(propertySupplier.invoke(DyeColor.GREEN)) }
		_redBlock = registry.register("${baseName}_red") { Block(propertySupplier.invoke(DyeColor.RED)) }
		_blackBlock = registry.register("${baseName}_black") { Block(propertySupplier.invoke(DyeColor.BLACK)) }
	}

	fun registerItems(registry: DeferredRegister<Item>) {
		_whiteItem = registry.register("${baseName}_white") { BlockItem(_whiteBlock.get(), Item.Properties().tab(OverdriveThatMatters.INSTANCE.CREATIVE_TAB_DECORATIVE).stacksTo(64)) }
		_orangeItem = registry.register("${baseName}_orange") { BlockItem(_orangeBlock.get(), Item.Properties().tab(OverdriveThatMatters.INSTANCE.CREATIVE_TAB_DECORATIVE).stacksTo(64)) }
		_magentaItem = registry.register("${baseName}_magenta") { BlockItem(_magentaBlock.get(), Item.Properties().tab(OverdriveThatMatters.INSTANCE.CREATIVE_TAB_DECORATIVE).stacksTo(64)) }
		_lightBlueItem = registry.register("${baseName}_light_blue") { BlockItem(_lightBlueBlock.get(), Item.Properties().tab(OverdriveThatMatters.INSTANCE.CREATIVE_TAB_DECORATIVE).stacksTo(64)) }
		_yellowItem = registry.register("${baseName}_yellow") { BlockItem(_yellowBlock.get(), Item.Properties().tab(OverdriveThatMatters.INSTANCE.CREATIVE_TAB_DECORATIVE).stacksTo(64)) }
		_limeItem = registry.register("${baseName}_lime") { BlockItem(_limeBlock.get(), Item.Properties().tab(OverdriveThatMatters.INSTANCE.CREATIVE_TAB_DECORATIVE).stacksTo(64)) }
		_pinkItem = registry.register("${baseName}_pink") { BlockItem(_pinkBlock.get(), Item.Properties().tab(OverdriveThatMatters.INSTANCE.CREATIVE_TAB_DECORATIVE).stacksTo(64)) }
		_grayItem = registry.register("${baseName}_gray") { BlockItem(_grayBlock.get(), Item.Properties().tab(OverdriveThatMatters.INSTANCE.CREATIVE_TAB_DECORATIVE).stacksTo(64)) }
		_lightGrayItem = registry.register("${baseName}_light_gray") { BlockItem(_lightGrayBlock.get(), Item.Properties().tab(OverdriveThatMatters.INSTANCE.CREATIVE_TAB_DECORATIVE).stacksTo(64)) }
		_cyanItem = registry.register("${baseName}_cyan") { BlockItem(_cyanBlock.get(), Item.Properties().tab(OverdriveThatMatters.INSTANCE.CREATIVE_TAB_DECORATIVE).stacksTo(64)) }
		_purpleItem = registry.register("${baseName}_purple") { BlockItem(_purpleBlock.get(), Item.Properties().tab(OverdriveThatMatters.INSTANCE.CREATIVE_TAB_DECORATIVE).stacksTo(64)) }
		_blueItem = registry.register("${baseName}_blue") { BlockItem(_blueBlock.get(), Item.Properties().tab(OverdriveThatMatters.INSTANCE.CREATIVE_TAB_DECORATIVE).stacksTo(64)) }
		_brownItem = registry.register("${baseName}_brown") { BlockItem(_brownBlock.get(), Item.Properties().tab(OverdriveThatMatters.INSTANCE.CREATIVE_TAB_DECORATIVE).stacksTo(64)) }
		_greenItem = registry.register("${baseName}_green") { BlockItem(_greenBlock.get(), Item.Properties().tab(OverdriveThatMatters.INSTANCE.CREATIVE_TAB_DECORATIVE).stacksTo(64)) }
		_redItem = registry.register("${baseName}_red") { BlockItem(_redBlock.get(), Item.Properties().tab(OverdriveThatMatters.INSTANCE.CREATIVE_TAB_DECORATIVE).stacksTo(64)) }
		_blackItem = registry.register("${baseName}_black") { BlockItem(_blackBlock.get(), Item.Properties().tab(OverdriveThatMatters.INSTANCE.CREATIVE_TAB_DECORATIVE).stacksTo(64)) }
	}
}

object MRegistry {
	private val features = RegistryDelegate<IForgeRegistry<AndroidFeatureType<*>>>("ANDROID_FEATURES")
	private val research = RegistryDelegate<IForgeRegistry<AndroidResearchType<*>>>("ANDROID_RESEARCH")

	val ANDROID_FEATURES get() = features.get() as ForgeRegistry<AndroidFeatureType<*>>
	val ANDROID_RESEARCH get() = research.get() as ForgeRegistry<AndroidResearchType<*>>

	val TRITANIUM_BLOCK = ColoredDecorativeBlock(MNames.TRITANIUM_BLOCK) {
		BlockBehaviour.Properties.of(Material.METAL, it.materialColor)
			.sound(SoundType.BASALT)
			.requiresCorrectToolForDrops()
			.explosionResistance(80f)
			.strength(4f)
	}

	val FLOOR_TILES = ColoredDecorativeBlock(MNames.FLOOR_TILES) {
		BlockBehaviour.Properties.of(Material.STONE, it.materialColor)
			.sound(SoundType.STONE)
			.requiresCorrectToolForDrops()
			.strength(3f)
	}

	private fun register(event: NewRegistryEvent) {
		features.write(RegistryBuilder<AndroidFeatureType<*>>().let {
			it.setName(ResourceLocation(OverdriveThatMatters.MOD_ID, "android_features"))
			it.type = AndroidFeatureType::class.java
			event.create(it)
		})

		research.write(RegistryBuilder<AndroidResearchType<*>>().let {
			it.setName(ResourceLocation(OverdriveThatMatters.MOD_ID, "android_research"))
			it.type = AndroidResearchType::class.java
			event.create(it)
		})
	}

	val DAMAGE_BECOME_ANDROID = ImmutableDamageSource(DamageSource("otm_become_android").bypassArmor().bypassInvul().bypassMagic())
	val DAMAGE_BECOME_HUMANE = ImmutableDamageSource(DamageSource("otm_become_humane").bypassArmor().bypassInvul().bypassMagic())
	val DAMAGE_EVENT_HORIZON = ImmutableDamageSource(DamageSource("otm_event_horizon").bypassMagic().bypassArmor())
	val DAMAGE_HAWKING_RADIATION = ImmutableDamageSource(DamageSource("otm_hawking_radiation"))
	const val DAMAGE_EMP_NAME = "otm_emp"
	const val DAMAGE_PLASMA_NAME = "otm_plasma"
	val DAMAGE_EMP: DamageSource = ImmutableDamageSource(EMPDamageSource())

	init {
		// DAMAGE_HAWKING_RADIATION.bypassMagic().bypassArmor();
	}

	val CRATE_RED = CrateProperties(MaterialColor.COLOR_RED, "crate_red")
	val CRATE_BLUE = CrateProperties(MaterialColor.COLOR_BLUE, "crate_blue")
	val CRATE_YELLOW = CrateProperties(MaterialColor.COLOR_YELLOW, "crate_yellow")
	val CRATE_GREEN = CrateProperties(MaterialColor.COLOR_GREEN, "crate_green")
	val CRATE_BLACK = CrateProperties(MaterialColor.COLOR_BLACK, "crate_black")
	val CRATE_PINK = CrateProperties(MaterialColor.COLOR_PINK, "crate_pink")
	val CRATE_PURPLE = CrateProperties(MaterialColor.COLOR_PURPLE, "crate_purple")

	val CRATE_LIST = listOf(
		CRATE_RED,
		CRATE_BLUE,
		CRATE_YELLOW,
		CRATE_GREEN,
		CRATE_BLACK,
		CRATE_PINK,
		CRATE_PURPLE,
	)

	val INDUSTRIAL_GLASS = IndustrialGlassProperties(null)
	val INDUSTRIAL_GLASS_WHITE = IndustrialGlassProperties(DyeColor.WHITE)
	val INDUSTRIAL_GLASS_ORANGE = IndustrialGlassProperties(DyeColor.ORANGE)
	val INDUSTRIAL_GLASS_MAGENTA = IndustrialGlassProperties(DyeColor.MAGENTA)
	val INDUSTRIAL_GLASS_LIGHT_BLUE = IndustrialGlassProperties(DyeColor.LIGHT_BLUE)
	val INDUSTRIAL_GLASS_YELLOW = IndustrialGlassProperties(DyeColor.YELLOW)
	val INDUSTRIAL_GLASS_LIME = IndustrialGlassProperties(DyeColor.LIME)
	val INDUSTRIAL_GLASS_PINK = IndustrialGlassProperties(DyeColor.PINK)
	val INDUSTRIAL_GLASS_GRAY = IndustrialGlassProperties(DyeColor.GRAY)
	val INDUSTRIAL_GLASS_LIGHT_GRAY = IndustrialGlassProperties(DyeColor.LIGHT_GRAY)
	val INDUSTRIAL_GLASS_CYAN = IndustrialGlassProperties(DyeColor.CYAN)
	val INDUSTRIAL_GLASS_PURPLE = IndustrialGlassProperties(DyeColor.PURPLE)
	val INDUSTRIAL_GLASS_BLUE = IndustrialGlassProperties(DyeColor.BLUE)
	val INDUSTRIAL_GLASS_BROWN = IndustrialGlassProperties(DyeColor.BROWN)
	val INDUSTRIAL_GLASS_GREEN = IndustrialGlassProperties(DyeColor.GREEN)
	val INDUSTRIAL_GLASS_RED = IndustrialGlassProperties(DyeColor.RED)
	val INDUSTRIAL_GLASS_BLACK = IndustrialGlassProperties(DyeColor.BLACK)

	val INDUSTRIAL_GLASS_LIST = listOf(
		INDUSTRIAL_GLASS,
		INDUSTRIAL_GLASS_WHITE,
		INDUSTRIAL_GLASS_ORANGE,
		INDUSTRIAL_GLASS_MAGENTA,
		INDUSTRIAL_GLASS_LIGHT_BLUE,
		INDUSTRIAL_GLASS_YELLOW,
		INDUSTRIAL_GLASS_LIME,
		INDUSTRIAL_GLASS_PINK,
		INDUSTRIAL_GLASS_GRAY,
		INDUSTRIAL_GLASS_LIGHT_GRAY,
		INDUSTRIAL_GLASS_CYAN,
		INDUSTRIAL_GLASS_PURPLE,
		INDUSTRIAL_GLASS_BLUE,
		INDUSTRIAL_GLASS_BROWN,
		INDUSTRIAL_GLASS_GREEN,
		INDUSTRIAL_GLASS_RED,
		INDUSTRIAL_GLASS_BLACK,
	)

	fun initialize(context: FMLJavaModLoadingContext) {
		context.modEventBus.addListener(this::register)
		context.modEventBus.register(MStats)
		context.modEventBus.register(LootModifiers)

		MBlocks.register()
		MBlockEntities.register()
		MEntityTypes.register()
		MMenus.register()
		MItems.register()
		AndroidFeatures.register()
		AndroidResearch.register()
		MSoundEvents.register()
	}
}
